/*
 * Copyright (c) 2006-2007 Extreme! Lab, Indiana University. All rights reserved.
 *
 * This software is open source. See the bottom of this file for the license.
 *
 * $Id: JythonClassLoader.java,v 1.5 2007/04/02 23:24:09 sshirasu Exp $
 */
package edu.indiana.extreme.xbaya.jython.runner;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLDecoder;
import java.security.AllPermission;
import java.security.CodeSigner;
import java.security.CodeSource;
import java.security.PermissionCollection;
import java.security.Permissions;
import java.security.SecureClassLoader;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import org.python.util.PythonInterpreter;

import xsul5.MLogger;
import edu.indiana.extreme.xbaya.XBayaRuntimeException;
import edu.indiana.extreme.xbaya.XBayaVersion;
import edu.indiana.extreme.xbaya.jython.lib.NotificationSender;
import edu.indiana.extreme.xbaya.util.IOUtil;

/**
 * This class loader loads jython related classes without counting on parent
 * class loader. This is because jython related classes use a lot of static
 * fields and cannot be used to invoke Jython scripts multiple times.
 * 
 * @author Satoshi Shirasuna
 */
public class JythonClassLoader extends SecureClassLoader {

    private final static MLogger logger = MLogger.getLogger();

    private ClassLoader parent;

    private Map<String, Class> classes = new HashMap<String, Class>();

    private URL jythonURL;

    private URL xbayaURL;

    private JarFile jythonJarFile;

    private JarFile xbayaJarFile;

    private File tmpJarDirectory;

    /**
     * Constructs a JythonClassLoader.
     * 
     * @param parent
     *            the parent class loader.
     * 
     * This has to be explicitly passed because WebStart applications use
     * user-level class loader. The default system loader cannot load classes in
     * the downloaded jar files.
     */
    public JythonClassLoader(ClassLoader parent) {
        super(parent);
        this.parent = parent;

        this.jythonURL = getBaseURL(PythonInterpreter.class);
        this.xbayaURL = getBaseURL(XBayaVersion.class);

        this.jythonJarFile = maybeGetJarFile(this.jythonURL);
        this.xbayaJarFile = maybeGetJarFile(this.xbayaURL);
    }

    /**
     * @return XBaya jar file.
     */
    public JarFile getXBayaJarFile() {
        return this.xbayaJarFile;
    }

    /**
     * Cleans up temporary files.
     */
    public void cleanUp() {
        if (this.tmpJarDirectory != null) {
            try {
                IOUtil.deleteDirectory(this.tmpJarDirectory);
            } catch (RuntimeException e) {
                logger.caught(e);
            }
        }
    }

    /**
     * @see java.lang.ClassLoader#loadClass(java.lang.String)
     */
    @Override
    public Class<?> loadClass(String className) throws ClassNotFoundException {
        return (loadClass(className, false));
    }

    /**
     * @see java.lang.ClassLoader#loadClass(java.lang.String, boolean)
     */
    @Override
    public synchronized Class<?> loadClass(String name, boolean resolveIt)
            throws ClassNotFoundException {

        // logger.entering(new Object[] { name, new Boolean(resolveIt) });

        Class klass = null;
        try {
            klass = findClass(name);
            // logger.finest("found from jython, klass: " + klass);
        } catch (ClassNotFoundException e) {
            try {
                klass = super.loadClass(name, false);
                // logger.finest("found from super, klass: " + klass);
            } catch (ClassNotFoundException e2) {
                klass = this.parent.loadClass(name);
                logger.finest("found from parent, klass: " + klass);
            }
        }

        if (resolveIt) {
            resolveClass(klass);
        }

        // logger.exiting(klass);
        return klass;
    }

    @Override
    protected Class<?> findClass(final String name)
            throws ClassNotFoundException {

        // logger.entering(new Object[] { name });

        Class klass;

        // Check if the class has been loaded by this class loader.
        klass = this.classes.get(name);
        if (klass != null) {
            return klass;
        }

        // If the class is python, load separatly. NotificationSender also uses
        // PyObject, so it has to be read from this class loader.
        // JythonOneTimeRunner also needs to be loaded separatly.
        if (name.startsWith("org.python.")) {
            klass = findClassFromURL(name, this.jythonURL, this.jythonJarFile);
        } else if (name.startsWith(NotificationSender.class.getPackage()
                .getName())
                || name.startsWith(JythonOneTimeRunnerImpl.class.getName())) {
            klass = findClassFromURL(name, this.xbayaURL, this.xbayaJarFile);
        }

        if (klass != null) {
            this.classes.put(name, klass);
            return klass;
        } else {
            throw new ClassNotFoundException();
        }
    }

    /**
     * @see java.security.SecureClassLoader#getPermissions(java.security.CodeSource)
     */
    @Override
    protected PermissionCollection getPermissions(CodeSource codesource) {
        logger.entering(new Object[] { codesource });
        // Grant all perission. This could be avoided if code signers were
        // extracted correctly.
        Permissions permissions = new Permissions();
        AllPermission permission = new AllPermission();
        permissions.add(permission);
        return permissions;
    }

    private URL getBaseURL(Class klass) {
        String path = klass.getName().replace('.', '/').concat(".class"); // /d/e/f.class
        URL classURL = this.parent.getResource(path);

        String jarURLString;
        if ("jar".equals(classURL.getProtocol())) {
            // classURL = jar:file/a/b/c.jar!/d/e/f.class
            // or jar:http://example.org/a/b/c.jar!/d/e/f.class
            String file = classURL.getFile();
            // file = file:/a/b/c.jar!d/e/f.class
            // or http://example.org/a/b/c.jar!d/e/f.class
            logger.finest("file: " + file);
            jarURLString = file.substring(0, file.lastIndexOf('!'));
            // jarURLString = file:/a/b/c.jar
            // or http://example.org/a/b/c.jar
        } else {
            // file:/a/b/c/d/e/f.class
            String file = classURL.getFile(); // /a/b/c/d/e/f.class
            int index = file.lastIndexOf(path);
            jarURLString = "file:" + file.substring(0, index); // /a/b/c/
        }
        try {
            URL jarURL = new URL(jarURLString);
            logger.exiting(jarURL);
            return jarURL;
        } catch (MalformedURLException e) {
            throw new XBayaRuntimeException(e);
        }
    }

    private JarFile maybeGetJarFile(URL url) {
        String path;
        try {
            path = URLDecoder.decode(url.getPath(), "UTF-8");
        } catch (UnsupportedEncodingException e) {
            throw new XBayaRuntimeException(e);
        }
        logger.finest("path: " + path);
        if (path.endsWith("/")) {
            // url = file:/a/b/c/
            // It's a local directory
            return null;
        } else if ("file".equals(url.getProtocol())) {
            // url = file:/a/b/c.jar
            // Jar file
            try {
                JarFile jarFile = new JarFile(path);
                return jarFile;
            } catch (IOException e) {
                throw new XBayaRuntimeException(e);
            }
        } else {
            // url = http://example.com/a/b/c.jar
            // A Jar file
            try {
                if (this.tmpJarDirectory == null) {
                    Date date = new Date();
                    SimpleDateFormat format = new SimpleDateFormat(
                            "yyyyMMdd-HHmmss-S");
                    String time = format.format(date);
                    String fileName = ".xbaya-jars-" + time;
                    String tmpdir = System.getProperty("java.io.tmpdir");
                    this.tmpJarDirectory = new File(tmpdir, fileName);
                    this.tmpJarDirectory.mkdir();
                }

                int i = path.lastIndexOf('/');
                File file = new File(this.tmpJarDirectory, path
                        .substring(i + 1));
                logger.finest("file: " + file);
                InputStream stream = url.openStream();
                IOUtil.writeToFile(stream, file);
                JarFile jarFile = new JarFile(file);
                return jarFile;
            } catch (IOException e) {
                throw new XBayaRuntimeException(e);
            }
        }
    }

    private Class findClassFromURL(String name, URL url, JarFile jarFile)
            throws ClassNotFoundException {
        // logger.entering(new Object[] { name, url, jarFile });

        String classPath = name.replace('.', '/').concat(".class");
        // logger.finest("classPath: " + classPath);

        try {
            byte[] classBytes;
            CodeSource codeSource = null;
            if (jarFile == null) {
                // It's a local directory
                String dirPath = URLDecoder.decode(url.getPath(), "UTF-8");
                File classFile = new File(dirPath, classPath);
                classBytes = IOUtil.readToByteArray(classFile);

            } else {
                // A Jar file
                JarEntry jarEntry = jarFile.getJarEntry(classPath);
                CodeSigner[] codeSigners = jarEntry.getCodeSigners();
                // logger.finest("codeSigners: " + codeSigners);
                if (codeSigners != null) {
                    // Somehow it's null.
                    for (CodeSigner signer : codeSigners) {
                        logger.finest("signer: " + signer);
                    }
                }
                codeSource = new CodeSource(this.xbayaURL, codeSigners);
                InputStream classInputStream = jarFile.getInputStream(jarEntry);
                classBytes = IOUtil.readToByteArray(classInputStream);
            }

            Class<?> klass = defineClass(name, classBytes, 0,
                    classBytes.length, codeSource);
            this.classes.put(name, klass);
            return klass;
        } catch (IOException e) {
            logger.caught(e);
            throw new ClassNotFoundException();
        }
    }

    /**
     * @see java.lang.ClassLoader#clearAssertionStatus()
     */
    @Override
    public synchronized void clearAssertionStatus() {
        logger.entering();
        super.clearAssertionStatus();
    }

    /**
     * @see java.lang.ClassLoader#definePackage(java.lang.String,
     *      java.lang.String, java.lang.String, java.lang.String,
     *      java.lang.String, java.lang.String, java.lang.String, java.net.URL)
     */
    @Override
    protected Package definePackage(String name, String specTitle,
            String specVersion, String specVendor, String implTitle,
            String implVersion, String implVendor, URL sealBase)
            throws IllegalArgumentException {
        logger.entering();
        return super.definePackage(name, specTitle, specVersion, specVendor,
                implTitle, implVersion, implVendor, sealBase);
    }

    /**
     * @see java.lang.ClassLoader#findLibrary(java.lang.String)
     */
    @Override
    protected String findLibrary(String libname) {
        logger.entering();
        return super.findLibrary(libname);
    }

    /**
     * @see java.lang.ClassLoader#findResource(java.lang.String)
     */
    @Override
    protected URL findResource(String name) {
        logger.entering();
        return super.findResource(name);
    }

    /**
     * @see java.lang.ClassLoader#findResources(java.lang.String)
     */
    @Override
    protected Enumeration<URL> findResources(String name) throws IOException {
        logger.entering();
        return super.findResources(name);
    }

    /**
     * @see java.lang.ClassLoader#getPackage(java.lang.String)
     */
    @Override
    protected Package getPackage(String name) {
        logger.entering();
        return super.getPackage(name);
    }

    /**
     * @see java.lang.ClassLoader#getPackages()
     */
    @Override
    protected Package[] getPackages() {
        logger.entering();
        return super.getPackages();
    }

    /**
     * @see java.lang.ClassLoader#getResource(java.lang.String)
     */
    @Override
    public URL getResource(String name) {
        logger.entering();
        return super.getResource(name);
    }

    /**
     * @see java.lang.ClassLoader#getResourceAsStream(java.lang.String)
     */
    @Override
    public InputStream getResourceAsStream(String name) {
        logger.entering();
        return super.getResourceAsStream(name);
    }

    /**
     * @see java.lang.ClassLoader#getResources(java.lang.String)
     */
    @Override
    public Enumeration<URL> getResources(String name) throws IOException {
        logger.entering();
        return super.getResources(name);
    }

    /**
     * @see java.lang.ClassLoader#setClassAssertionStatus(java.lang.String,
     *      boolean)
     */
    @Override
    public synchronized void setClassAssertionStatus(String className,
            boolean enabled) {
        logger.entering();
        super.setClassAssertionStatus(className, enabled);
    }

    /**
     * @see java.lang.ClassLoader#setDefaultAssertionStatus(boolean)
     */
    @Override
    public synchronized void setDefaultAssertionStatus(boolean enabled) {
        logger.entering();
        super.setDefaultAssertionStatus(enabled);
    }

    /**
     * @see java.lang.ClassLoader#setPackageAssertionStatus(java.lang.String,
     *      boolean)
     */
    @Override
    public synchronized void setPackageAssertionStatus(String packageName,
            boolean enabled) {
        logger.entering();
        super.setPackageAssertionStatus(packageName, enabled);
    }

}

/*
 * Indiana University Extreme! Lab Software License, Version 1.2
 * 
 * Copyright (c) 2006-2007 The Trustees of Indiana University. All rights
 * reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 * 
 * 2) All redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 * 
 * 3) Any documentation included with all redistributions must include the
 * following acknowledgement:
 * 
 * "This product includes software developed by the Indiana University Extreme!
 * Lab. For further information please visit http://www.extreme.indiana.edu/"
 * 
 * Alternatively, this acknowledgment may appear in the software itself, and
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name "Indiana University" or "Indiana University Extreme! Lab" shall
 * not be used to endorse or promote products derived from this software without
 * prior written permission from Indiana University. For written permission,
 * please contact http://www.extreme.indiana.edu/.
 * 
 * 5) Products derived from this software may not use "Indiana University" name
 * nor may "Indiana University" appear in their name, without prior written
 * permission of the Indiana University.
 * 
 * Indiana University provides no reassurances that the source code provided
 * does not infringe the patent or any other intellectual property rights of any
 * other entity. Indiana University disclaims any liability to any recipient for
 * claims brought by any other entity based on infringement of intellectual
 * property rights or otherwise.
 * 
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS.
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS",
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION
 * GENERATED USING SOFTWARE.
 */
